@@ -36,7 +36,7 @@ ensure_bitarray(PyObject *obj)
 
 /* ensure object is a bitarray of given length */
 static int
-ensure_ba_of_length(PyObject *a, const Py_ssize_t n)
+ensure_ba_of_length(PyObject *a, Py_ssize_t n)
 {
     if (ensure_bitarray(a) < 0)
         return -1;
@@ -55,6 +55,7 @@ static Py_ssize_t
 count_to_n(bitarrayobject *a, Py_ssize_t n, int vi)
 {
     const Py_ssize_t nbits = a->nbits;
+    unsigned char *ucbuff = (unsigned char *) a->ob_item;
     Py_ssize_t i = 0;        /* index */
     Py_ssize_t j = 0;        /* total count up to index */
     Py_ssize_t block_start, block_stop, k, m;
@@ -72,7 +73,7 @@ count_to_n(bitarrayobject *a, Py_ssize_t n, int vi)
         block_stop = block_start + (BLOCK_BITS >> 3);
         assert(block_stop <= Py_SIZE(a));
         for (k = block_start; k < block_stop; k++)
-            m += bitcount_lookup[(unsigned char) a->ob_item[k]];
+            m += bitcount_lookup[ucbuff[k]];
         if (!vi)
             m = BLOCK_BITS - m;
         if (j + m >= n)
@@ -85,7 +86,7 @@ count_to_n(bitarrayobject *a, Py_ssize_t n, int vi)
     while (i + 8 < nbits) {
         k = i >> 3;
         assert(k < Py_SIZE(a));
-        m = bitcount_lookup[(unsigned char) a->ob_item[k]];
+        m = bitcount_lookup[ucbuff[k]];
         if (!vi)
             m = 8 - m;
         if (j + m >= n)
@@ -160,31 +161,31 @@ find_last(bitarrayobject *self, int vi, Py_ssize_t a, Py_ssize_t b)
     /* the logic here is the same as in find_bit() in _bitarray.c */
 #ifdef PY_UINT64_T
     if (n > 64) {
-        const Py_ssize_t word_a = (a + 63) / 64;
-        const Py_ssize_t word_b = b / 64;
-        const PY_UINT64_T w = vi ? 0 : ~0;
+        Py_ssize_t word_a = (a + 63) / 64;
+        Py_ssize_t word_b = b / 64;
+        PY_UINT64_T *wbuff = (PY_UINT64_T *) self->ob_item, w = vi ? 0 : ~0;
 
         if ((res = find_last(self, vi, 64 * word_b, b)) >= 0)
             return res;
 
         for (i = word_b - 1; i >= word_a; i--) {  /* skip uint64 words */
-            if (w ^ ((PY_UINT64_T *) self->ob_item)[i])
+            if (w ^ wbuff[i])
                 return find_last(self, vi, 64 * i, 64 * i + 64);
         }
         return find_last(self, vi, a, 64 * word_a);
     }
 #endif
     if (n > 8) {
-        const Py_ssize_t byte_a = BYTES(a);
-        const Py_ssize_t byte_b = b / 8;
-        const char c = vi ? 0 : ~0;
+        Py_ssize_t byte_a = BYTES(a);
+        Py_ssize_t byte_b = b / 8;
+        char *buff = self->ob_item, c = vi ? 0 : ~0;
 
         if ((res = find_last(self, vi, 8 * byte_b, b)) >= 0)
             return res;
 
         for (i = byte_b - 1; i >= byte_a; i--) {  /* skip bytes */
             assert_byte_in_range(self, i);
-            if (c ^ self->ob_item[i])
+            if (c ^ buff[i])
                 return find_last(self, vi, 8 * i, 8 * i + 8);
         }
         return find_last(self, vi, a, 8 * byte_a);
@@ -212,8 +213,7 @@ r_index(PyObject *module, PyObject *args)
         return NULL;
 
 #define aa  ((bitarrayobject *) a)
-    normalize_index(aa->nbits, 1, &start);
-    normalize_index(aa->nbits, 1, &stop);
+    adjust_indices(aa->nbits, &start, &stop, 1);
     res = find_last(aa, vi, start, stop);
 #undef aa
     if (res < 0)
@@ -269,6 +269,7 @@ binary_function(PyObject *args, enum kernel_type kern, const char *format)
 {
     Py_ssize_t res = 0, s, i;
     PyObject *a, *b;
+    char *buff_a, *buff_b;
     unsigned char c;
     int r;
 
@@ -289,13 +290,15 @@ binary_function(PyObject *args, enum kernel_type kern, const char *format)
                         "bitarrays of equal endianness expected");
         return NULL;
     }
+    buff_a = aa->ob_item;
+    buff_b = bb->ob_item;
     s = aa->nbits / 8;       /* number of whole bytes in buffer */
     r = aa->nbits % 8;       /* remaining bits  */
 
     switch (kern) {
     case KERN_cand:
         for (i = 0; i < s; i++) {
-            c = aa->ob_item[i] & bb->ob_item[i];
+            c = buff_a[i] & buff_b[i];
             res += bitcount_lookup[c];
         }
         if (r) {
@@ -306,7 +309,7 @@ binary_function(PyObject *args, enum kernel_type kern, const char *format)
 
     case KERN_cor:
         for (i = 0; i < s; i++) {
-            c = aa->ob_item[i] | bb->ob_item[i];
+            c = buff_a[i] | buff_b[i];
             res += bitcount_lookup[c];
         }
         if (r) {
@@ -317,7 +320,7 @@ binary_function(PyObject *args, enum kernel_type kern, const char *format)
 
     case KERN_cxor:
         for (i = 0; i < s; i++) {
-            c = aa->ob_item[i] ^ bb->ob_item[i];
+            c = buff_a[i] ^ buff_b[i];
             res += bitcount_lookup[c];
         }
         if (r) {
@@ -328,7 +331,7 @@ binary_function(PyObject *args, enum kernel_type kern, const char *format)
 
     case KERN_subset:
         for (i = 0; i < s; i++) {
-            if ((aa->ob_item[i] & bb->ob_item[i]) != aa->ob_item[i])
+            if ((buff_a[i] & buff_b[i]) != buff_a[i])
                 Py_RETURN_FALSE;
         }
         if (r) {
@@ -608,7 +611,7 @@ ba2base(PyObject *module, PyObject *args)
         int j, x = 0;
 
         for (j = 0; j < m; j++) {
-            const int k = le ? j : (m - j - 1);
+            int k = le ? j : (m - j - 1);
             x |= getbit(aa, i * m + k) << j;
         }
         str[i] = alphabet[x];
@@ -665,7 +668,7 @@ base2ba(PyObject *module, PyObject *args)
                                 "base %d, got '%c' (0x%02x)", n, c, c);
         }
         for (j = 0; j < m; j++) {
-            const int k = le ? j : (m - j - 1);
+            int k = le ? j : (m - j - 1);
             setbit(aa, i * m + k, d & (1 << j));
         }
     }
@@ -736,7 +739,7 @@ vl_decode(PyObject *module, PyObject *args)
             b = (unsigned char) *PyBytes_AS_STRING(item);
 #endif
         else {
-            PyErr_Format(PyExc_TypeError, "int (byte) iterator expected, "
+            PyErr_Format(PyExc_TypeError, "int iterator expected, "
                          "got '%s' element", Py_TYPE(item)->tp_name);
             Py_DECREF(item);
             return NULL;
@@ -806,7 +809,7 @@ vl_encode(PyObject *module, PyObject *a)
         str[0] |= (0x08 >> i) * getbit(aa, i);
 
     for (i = 4; i < aa->nbits; i++) {
-        const int k = (i - 4) % 7;
+        int k = (i - 4) % 7;
 
         if (k == 0) {
             j++;
@@ -827,6 +830,215 @@ Return variable length binary representation of bitarray.\n\
 This representation is useful for efficiently storing small bitarray\n\
 in a binary stream.  Use `vl_decode()` for decoding.");
 
+/* ----------------------- canonical Huffman decoder ------------------- */
+
+/*
+   The decode iterator object includes the Huffman code decoding tables:
+   - count[1..MAXBITS] is the number of symbols of each length, which for a
+     canonical code are stepped through in order.  count[0] is not used.
+   - symbol is a Python sequence of the symbols in canonical order
+     where the number of entries is the sum of the counts in count[].
+ */
+#define MAXBITS  31                  /* maximum bits in a code */
+
+typedef struct {
+    PyObject_HEAD
+    bitarrayobject *array;           /* bitarray we're decoding */
+    Py_ssize_t index;                /* current index in bitarray */
+    int count[MAXBITS + 1];          /* number of symbols of each length */
+    PyObject *symbol;                /* canonical ordered symbols */
+} chdi_obj;                          /* canonical Huffman decode iterator */
+
+static PyTypeObject CHDI_Type;
+
+/* set elements in count (from seq) and return their sum, or -1 on error */
+static Py_ssize_t
+set_count(int *count, PyObject *sequence)
+{
+    Py_ssize_t n, c, res = 0;
+    int i;
+
+    n = PySequence_Size(sequence);
+    if (n < 0)
+        return -1;
+
+    if (n > MAXBITS) {
+        PyErr_Format(PyExc_ValueError, "len(count) cannot be larger than %d",
+                     MAXBITS);
+        return -1;
+    }
+
+    for (i = 1; i <= MAXBITS; i++) {
+        c = 0;
+        if (i < n) {
+            PyObject *item = PySequence_GetItem(sequence, i);
+            Py_ssize_t maxcount = ((Py_ssize_t) 1) << i;
+
+            if (item == NULL)
+                return -1;
+            c = PyNumber_AsSsize_t(item, PyExc_OverflowError);
+            Py_DECREF(item);
+            if (c == -1 && PyErr_Occurred())
+                return -1;
+            if (c < 0 || c > maxcount) {
+                PyErr_Format(PyExc_ValueError, "count[%d] cannot be negative"
+                             " or larger than %zd, got %zd", i, maxcount, c);
+                return -1;
+            }
+        }
+        count[i] = (int) c;
+        res += c;
+    }
+    return res;
+}
+
+/* create a new initialized canonical Huffman decode iterator object */
+static PyObject *
+chdi_new(PyObject *module, PyObject *args)
+{
+    PyObject *a, *count, *symbol;
+    Py_ssize_t count_sum;
+    chdi_obj *it;       /* iterator object to be returned */
+
+    if (!PyArg_ParseTuple(args, "OOO:count_n", &a, &count, &symbol))
+        return NULL;
+    if (ensure_bitarray(a) < 0)
+        return NULL;
+    if (!PySequence_Check(count))
+        return PyErr_Format(PyExc_TypeError, "count expected to be sequence, "
+                            "got '%s'", Py_TYPE(count)->tp_name);
+
+    symbol = PySequence_Fast(symbol, "symbol not iterable");
+    if (symbol == NULL)
+        return NULL;
+
+    it = PyObject_GC_New(chdi_obj, &CHDI_Type);
+    if (it == NULL)
+        goto error;
+
+    if ((count_sum = set_count(it->count, count)) < 0)
+        goto error;
+
+    if (count_sum != PySequence_Size(symbol)) {
+        PyErr_Format(PyExc_ValueError, "sum(count) = %zd, but len(symbol) "
+                     "= %zd", count_sum, PySequence_Size(symbol));
+        goto error;
+    }
+    Py_INCREF(a);
+    it->array = (bitarrayobject *) a;
+    it->index = 0;
+    /* PySequence_Fast() returns a new reference, so no Py_INCREF here */
+    it->symbol = symbol;
+
+    PyObject_GC_Track(it);
+    return (PyObject *) it;
+
+ error:
+    it->array = NULL;
+    Py_XDECREF(symbol);
+    it->symbol = NULL;
+    Py_DECREF((PyObject *) it);
+    return NULL;
+}
+
+PyDoc_STRVAR(chdi_doc,
+"canonical_decode(bitarray, count, symbol, /) -> iterator\n\
+\n\
+Decode bitarray using canonical Huffman decoding tables\n\
+where `count` is a sequence containing the number of symbols of each length\n\
+and `symbol` is a sequence of symbols in canonical order.");
+
+/* This function is based on the function decode() in:
+   https://github.com/madler/zlib/blob/master/contrib/puff/puff.c */
+static PyObject *
+chdi_next(chdi_obj *it)
+{
+    Py_ssize_t nbits = it->array->nbits;
+    int len;    /* current number of bits in code */
+    int code;   /* current code (of len bits) */
+    int first;  /* first code of length len */
+    int count;  /* number of codes of length len */
+    int index;  /* index of first code of length len in symbol list */
+
+    if (it->index >= nbits)           /* no bits - stop iteration */
+        return NULL;
+
+    code = first = index = 0;
+    for (len = 1; len <= MAXBITS; len++) {
+        code |= getbit(it->array, it->index++);
+        count = it->count[len];
+        assert(code - first >= 0);
+        if (code - first < count) {   /* if length len, return symbol */
+            return PySequence_ITEM(it->symbol, index + (code - first));
+        }
+        index += count;               /* else update for next length */
+        first += count;
+        first <<= 1;
+        code <<= 1;
+
+        if (it->index >= nbits && len != MAXBITS) {
+            PyErr_SetString(PyExc_ValueError, "reached end of bitarray");
+            return NULL;
+        }
+    }
+    PyErr_SetString(PyExc_ValueError, "ran out of codes");
+    return NULL;
+}
+
+static void
+chdi_dealloc(chdi_obj *it)
+{
+    PyObject_GC_UnTrack(it);
+    Py_XDECREF(it->array);
+    Py_XDECREF(it->symbol);
+    PyObject_GC_Del(it);
+}
+
+static int
+chdi_traverse(chdi_obj *it, visitproc visit, void *arg)
+{
+    Py_VISIT(it->array);
+    Py_VISIT(it->symbol);
+    return 0;
+}
+
+static PyTypeObject CHDI_Type = {
+#ifdef IS_PY3K
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
+    PyObject_HEAD_INIT(NULL)
+    0,                                        /* ob_size */
+#endif
+    "bitarray.util.canonical_decodeiter",     /* tp_name */
+    sizeof(chdi_obj),                         /* tp_basicsize */
+    0,                                        /* tp_itemsize */
+    /* methods */
+    (destructor) chdi_dealloc,                /* tp_dealloc */
+    0,                                        /* tp_print */
+    0,                                        /* tp_getattr */
+    0,                                        /* tp_setattr */
+    0,                                        /* tp_compare */
+    0,                                        /* tp_repr */
+    0,                                        /* tp_as_number */
+    0,                                        /* tp_as_sequence */
+    0,                                        /* tp_as_mapping */
+    0,                                        /* tp_hash */
+    0,                                        /* tp_call */
+    0,                                        /* tp_str */
+    PyObject_GenericGetAttr,                  /* tp_getattro */
+    0,                                        /* tp_setattro */
+    0,                                        /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,  /* tp_flags */
+    0,                                        /* tp_doc */
+    (traverseproc) chdi_traverse,             /* tp_traverse */
+    0,                                        /* tp_clear */
+    0,                                        /* tp_richcompare */
+    0,                                        /* tp_weaklistoffset */
+    PyObject_SelfIter,                        /* tp_iter */
+    (iternextfunc) chdi_next,                 /* tp_iternext */
+    0,                                        /* tp_methods */
+};
+
 /* --------------------------------------------------------------------- */
 
 /* Set bitarray_type_obj (bato).  This function must be called before any
@@ -853,6 +1065,8 @@ static PyMethodDef module_functions[] = {
     {"_base2ba",  (PyCFunction) base2ba,   METH_VARARGS, 0},
     {"vl_encode", (PyCFunction) vl_encode, METH_O,       vl_encode_doc},
     {"_vl_decode",(PyCFunction) vl_decode, METH_VARARGS, 0},
+    {"canonical_decode",
+                  (PyCFunction) chdi_new,  METH_VARARGS, chdi_doc},
     {"_set_bato", (PyCFunction) set_bato,  METH_O,       0},
     {NULL,        NULL}  /* sentinel */
 };
@@ -876,12 +1090,22 @@ init_util(void)
 
 #ifdef IS_PY3K
     m = PyModule_Create(&moduledef);
-    if (m == NULL)
-        return NULL;
-    return m;
 #else
     m = Py_InitModule3("_util", module_functions, 0);
+#endif
     if (m == NULL)
-        return;
+        goto error;
+
+    if (PyType_Ready(&CHDI_Type) < 0)
+        goto error;
+    Py_SET_TYPE(&CHDI_Type, &PyType_Type);
+
+#ifdef IS_PY3K
+    return m;
+ error:
+    return NULL;
+#else
+ error:
+    return;
 #endif
 }
