@@ -4,7 +4,7 @@
 
    Author: Ilan Schnell
 */
-#define BITARRAY_VERSION  "2.4.1"
+#define BITARRAY_VERSION  "2.5.0"
 
 #ifdef STDC_HEADERS
 #include <stddef.h>
@@ -128,7 +128,7 @@ zeroed_last_byte(bitarrayobject *self)
 {
     const int r = self->nbits % 8;     /* index into mask table */
 
-    assert(r % 8 > 0);
+    assert(r > 0);
     assert_nbits(self);
     return ones_table[IS_BE(self)][r] & self->ob_item[Py_SIZE(self) - 1];
 }
@@ -157,12 +157,9 @@ static const unsigned char bitcount_lookup[256] = {
 #undef B6
 };
 
-/* normalize index (which may be negative), such that:
-    0 <= i <= length        (for step > 0)
-   -1 <= i <= length - 1    (for step < 0)
-*/
+/* adjust index a manner consistent with the handling of normal slices */
 static inline void
-normalize_index(Py_ssize_t length, Py_ssize_t step, Py_ssize_t *i)
+adjust_index(Py_ssize_t length, Py_ssize_t *i, Py_ssize_t step)
 {
     if (*i < 0) {
         *i += length;
@@ -174,6 +171,52 @@ normalize_index(Py_ssize_t length, Py_ssize_t step, Py_ssize_t *i)
     }
 }
 
+/* same as PySlice_AdjustIndices() which was introduced in Python 3.6.1 */
+static inline Py_ssize_t
+adjust_indices(Py_ssize_t length, Py_ssize_t *start, Py_ssize_t *stop,
+               Py_ssize_t step)
+{
+#if PY_VERSION_HEX > 0x03060100
+    return PySlice_AdjustIndices(length, start, stop, step);
+#else
+    assert(step != 0);
+    adjust_index(length, start, step);
+    adjust_index(length, stop, step);
+    /*
+      a / b does integer division.  If either a or b is negative, the result
+      depends on the compiler (rounding can go toward 0 or negative infinity).
+      Therefore, we are careful that both a and b are always positive.
+    */
+    if (step < 0) {
+        if (*stop < *start)
+            return (*start - *stop - 1) / (-step) + 1;
+    }
+    else {
+        if (*start < *stop)
+            return (*stop - *start - 1) / step + 1;
+    }
+    return 0;
+#endif
+}
+
+/* adjust slice parameters such that step is always positive; produces
+   simpler loops over elements when their order is irrelevant */
+static inline void
+adjust_step_positive(Py_ssize_t slicelength,
+                     Py_ssize_t *start, Py_ssize_t *stop, Py_ssize_t *step)
+{
+    if (*step < 0) {
+        *stop = *start + 1;
+        *start = *stop + *step * (slicelength - 1) - 1;
+        *step = -(*step);
+    }
+    assert(*start >= 0 && *stop >= 0 && *step > 0 && slicelength >= 0);
+    /* slicelength == 0 implies stop <= start */
+    assert(slicelength != 0 || *stop <= *start);
+    /* step == 1 and slicelength != 0 implies stop - start == slicelength */
+    assert(*step != 1 || slicelength == 0 || *stop - *start == slicelength);
+}
+
 /* Interpret a PyObject (usually PyLong or PyBool) as a bit, return 0 or 1.
    On error, return -1 and set error message. */
 static inline int
